@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,8 +19,12 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
-import renpy.display
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *
+
+from typing import Optional
+
+import renpy
 import time
 import collections
 
@@ -178,7 +182,7 @@ def cache_get(screen, args, kwargs):
     for sc in pc:
 
         # Reuse w/ same arguments.
-        if sc.args == args and sc.args == kwargs:
+        if sc.args == args and sc.kwargs == kwargs:
             pc.remove(sc)
             break
     else:
@@ -188,9 +192,9 @@ def cache_get(screen, args, kwargs):
 
     return sc.cache
 
-
 # Screens #####################################################################
 
+
 class Screen(renpy.object.Object):
     """
     A screen is a collection of widgets that are displayed together.
@@ -198,6 +202,7 @@ class Screen(renpy.object.Object):
     """
 
     sensitive = "True"
+    roll_forward = None
 
     def __init__(self,
                  name,
@@ -210,7 +215,8 @@ class Screen(renpy.object.Object):
                  parameters=False,
                  location=None,
                  layer="screens",
-                 sensitive="True"):
+                 sensitive="True",
+                 roll_forward=None):
 
         # The name of this screen.
         if isinstance(name, basestring):
@@ -230,7 +236,7 @@ class Screen(renpy.object.Object):
 
         # If this is a SL2 screen, the SLScreen node at the root of this
         # screen.
-        if isinstance(function, renpy.sl2.slast.SLScreen):  # @UndefinedVariable
+        if isinstance(function, renpy.sl2.slast.SLScreen): # @UndefinedVariable
             self.ast = function
         else:
             self.ast = None
@@ -262,6 +268,11 @@ class Screen(renpy.object.Object):
         # Is this screen sensitive? An expression.
         self.sensitive = sensitive
 
+        # Does the screen participate in roll_forward when used with the
+        # call screen statement? True for yes, False for no, None for
+        # config.call_screen_roll_forward.
+        self.roll_forward = roll_forward
+
         global prepared
         global analyzed
 
@@ -270,11 +281,11 @@ class Screen(renpy.object.Object):
 
 
 # Phases we can be in.
-PREDICT = 0  # Predicting the screen before it is shown.
-SHOW = 1    # Showing the screen for the first time.
-UPDATE = 2  # Showing the screen for the second and later times.
-HIDE = 3    # After the screen has been hid with "hide screen" (or the end of call screen).
-OLD = 4     # A copy of the screen in the old side of a transition.
+PREDICT = 0 # Predicting the screen before it is shown.
+SHOW = 1 # Showing the screen for the first time.
+UPDATE = 2 # Showing the screen for the second and later times.
+HIDE = 3 # After the screen has been hid with "hide screen" (or the end of call screen).
+OLD = 4 # A copy of the screen in the old side of a transition.
 
 phase_name = [
     "PREDICT",
@@ -297,6 +308,7 @@ class ScreenDisplayable(renpy.display.layout.Container):
         'children',
         'transforms',
         'widgets',
+        'base_widgets',
         'old_widgets',
         'hidden_widgets',
         'old_transforms',
@@ -316,8 +328,10 @@ class ScreenDisplayable(renpy.display.layout.Container):
         self.children = [ ]
         self.transforms = { }
         self.widgets = { }
+        self.base_widgets = { }
         self.old_widgets = None
         self.old_transforms = None
+        self.old_transfers = False
         self.hidden_widgets = { }
         self.cache = { }
         self.phase = UPDATE
@@ -349,10 +363,10 @@ class ScreenDisplayable(renpy.display.layout.Container):
 
         # The scope associated with this statement. This is passed in
         # as keyword arguments to the displayable.
-        self.scope = renpy.python.RevertableDict(scope)
+        self.scope = renpy.revertable.RevertableDict(scope)
 
         # The child associated with this screen.
-        self.child = None
+        self.child = None # type: renpy.display.layout.MultiBox|None
 
         # Widget properties given to this screen the last time it was
         # shown.
@@ -361,6 +375,9 @@ class ScreenDisplayable(renpy.display.layout.Container):
         # A map from name to the widget with that name.
         self.widgets = { }
 
+        # Same, but to the widget without considering _main.
+        self.base_widgets = { }
+
         # The persistent cache.
         self.cache = { }
 
@@ -420,8 +437,12 @@ class ScreenDisplayable(renpy.display.layout.Container):
         # The lifecycle phase we are in - one of PREDICT, SHOW, UPDATE, or HIDE.
         self.phase = PREDICT
 
-    def __unicode__(self):
-        return "Screen {}".format(" ".join(self.screen_name))
+    @property
+    def name(self):
+        return " ".join(self.screen_name)
+
+    def _repr_info(self):
+        return self.name
 
     def visit(self):
         return [ self.child ]
@@ -431,7 +452,8 @@ class ScreenDisplayable(renpy.display.layout.Container):
 
         try:
             push_current_screen(self)
-            self.child.visit_all(callback, seen=None)
+            if self.child is not None:
+                self.child.visit_all(callback, seen=None)
         finally:
             pop_current_screen()
 
@@ -447,6 +469,9 @@ class ScreenDisplayable(renpy.display.layout.Container):
 
         hiding = (self.phase == OLD) or (self.phase == HIDE)
 
+        if self.modal and not callable(self.modal):
+            renpy.display.focus.mark_modal()
+
         try:
             push_current_screen(self)
 
@@ -455,10 +480,12 @@ class ScreenDisplayable(renpy.display.layout.Container):
         finally:
             pop_current_screen()
 
+
     def copy(self):
         rv = ScreenDisplayable(self.screen, self.tag, self.layer, self.widget_properties, self.scope, **self.properties)
         rv.transforms = self.transforms.copy()
         rv.widgets = self.widgets.copy()
+        rv.base_widgets = self.widgets.copy()
         rv.old_transfers = True
         rv.child = self.child
 
@@ -559,13 +586,13 @@ class ScreenDisplayable(renpy.display.layout.Container):
         updated_screens.add(self)
 
         if self.screen is None:
-            self.child = renpy.display.layout.Null()
+            self.child = renpy.display.layout.MultiBox(layout='fixed')
             return { }
 
         # Do not update if restarting or hiding.
         if self.restarting or (self.phase == HIDE) or (self.phase == OLD):
             if not self.child:
-                self.child = renpy.display.layout.Null()
+                self.child = renpy.display.layout.MultiBox(layout='fixed')
 
             return self.widgets
 
@@ -599,6 +626,7 @@ class ScreenDisplayable(renpy.display.layout.Container):
         self.old_widgets = self.widgets
         self.old_transforms = self.transforms
         self.widgets = { }
+        self.base_widgets = { }
         self.transforms = { }
 
         push_current_screen(self)
@@ -627,7 +655,8 @@ class ScreenDisplayable(renpy.display.layout.Container):
             renpy.ui.close()
 
         finally:
-            del self.scope["_scope"]
+            # Safe removal as to not reraise another exception and lose the last one
+            self.scope.pop("_scope", None)
 
             renpy.ui.screen = old_ui_screen
             pop_current_screen()
@@ -655,7 +684,7 @@ class ScreenDisplayable(renpy.display.layout.Container):
             end = time.time()
 
             if self.profile.time:
-                profile_log.write("* %.2f ms", 1000 * (end - start))
+                profile_log.write("* %.2f ms", 1000 * (end - start)) # type: ignore
 
             if self.profile.debug:
                 profile_log.write("\n")
@@ -718,16 +747,30 @@ class ScreenDisplayable(renpy.display.layout.Container):
         if rv is not None:
             return rv
 
-        if self.modal:
+        if renpy.display.layout.check_modal(self.modal, ev, x, y, None, None):
             raise renpy.display.layout.IgnoreLayers()
 
     def get_phase_name(self):
         return phase_name[self.phase]
 
+    def _tts(self):
+        if (self.phase == OLD) or (self.phase == HIDE):
+            return ""
+
+        if self.modal:
+            return renpy.display.tts.TTSDone(self._tts_common())
+        else:
+            return self._tts_common()
+
 
 # The name of the screen that is currently being displayed, or
 # None if no screen is being currently displayed.
-_current_screen = None
+_current_screen = None # type: renpy.display.screen.ScreenDisplayable|None
+
+if 0 == 1:
+    _current_screen = renpy.display.screen.ScreenDisplayable() # type: ignore # fake out typing.
+else:
+    _current_screen = None
 
 # The stack of old current screens.
 current_screen_stack = [ ]
@@ -781,13 +824,15 @@ def get_all_screen_variants(name):
     order.
     """
 
-    rv = [ ]
+    if isinstance(name, basestring):
+        name = tuple(name.split())
 
-    for k, v in screens.iteritems():
-        if k[0] == name:
-            rv.append((k[1], v))
+    name = name[0]
 
-    return rv
+    if name not in screens_by_name:
+        return [ ]
+
+    return list(screens_by_name[name].items())
 
 
 # Have all screens been analyzed?
@@ -862,7 +907,7 @@ def sort_screens():
         del reverse[name]
 
     # Store the use cycle for later reporting.
-    use_cycle = reverse.keys()
+    use_cycle = list(reverse.keys())
     use_cycle.sort()
 
     sorted_screens = rv
@@ -936,12 +981,12 @@ def prepare_screens():
         renpy.display.predict.predicting = old_predicting
 
     if renpy.config.developer and use_cycle:
-        raise Exception("The following screens use each other in a loop: " + ", ".join(use_cycle) +". This is not allowed.")
+        raise Exception("The following screens use each other in a loop: " + ", ".join(use_cycle) + ". This is not allowed.")
 
 
 def define_screen(*args, **kwargs):
     """
-    :doc: screens
+    :undocumented:
     :args: (name, function, modal="False", zorder="0", tag=None, variant=None)
 
     Defines a screen with `name`, which should be a string.
@@ -1002,8 +1047,8 @@ def get_screen(name, layer=None):
     """
     :doc: screens
 
-    Returns the ScreenDisplayable with the given `name` on layer. `name`
-    is first interpreted as a tag name, and then a screen name. If the
+    Returns the ScreenDisplayable with the given `name` on `layer`. `name`
+    is first interpreted as a tag name, and then as a screen name. If the
     screen is not showing, returns None.
 
     This can also take a list of names, in which case the first screen
@@ -1016,13 +1061,27 @@ def get_screen(name, layer=None):
         else:
             text "The say screen is hidden."
 
+    The ScreenDisplayable objects returned by this function have the
+    following documented fields:
+
+    .. attribute:: ScreenDisplayable.layer
+
+        The layer the screen is being displayed on.
+
+    .. attribute:: ScreenDisplayable.name
+
+        The name of the screen.
+
+    .. attribute:: ScreenDisplayable.zorder
+
+        The zorder the screen is being displayed at.
     """
 
     if layer is None:
         layer = get_screen_layer(name)
 
     if isinstance(name, basestring):
-        name = (name, )
+        name = (name,)
 
     sl = renpy.exports.scene_lists()
 
@@ -1034,7 +1093,7 @@ def get_screen(name, layer=None):
 
     for tag in name:
 
-        sd = sl.get_displayable_by_name(layer, (tag, ))
+        sd = sl.get_displayable_by_name(layer, (tag,))
         if sd is not None:
             return sd
 
@@ -1058,6 +1117,25 @@ def has_screen(name):
         return False
 
 
+def get_screen_roll_forward(screen_name):
+    """
+    Given a screeen name, determines if roll forward is enable for the
+    screen.
+    """
+
+    name = screen_name
+
+    if not isinstance(name, tuple):
+        name = tuple(name.split())
+
+    screen = get_screen_variant(name[0])
+
+    if screen is None:
+        raise Exception("Screen %s is not known.\n" % (name[0],))
+
+    return screen.roll_forward
+
+
 def show_screen(_screen_name, *_args, **kwargs):
     """
     :doc: screens
@@ -1162,7 +1240,7 @@ def predict_screen(_screen_name, *_args, **kwargs):
 
     name = _screen_name
 
-    if renpy.config.debug_image_cache:
+    if renpy.config.debug_prediction:
         renpy.display.ic_log.write("Predict screen %s", name)
 
     if not isinstance(name, tuple):
@@ -1173,6 +1251,9 @@ def predict_screen(_screen_name, *_args, **kwargs):
     if screen is None:
         return
 
+    if not screen.predict:
+        return
+
     if _layer is None:
         _layer = get_screen_layer(name)
 
@@ -1187,12 +1268,6 @@ def predict_screen(_screen_name, *_args, **kwargs):
 
     try:
 
-        if screen is None:
-            raise Exception("Screen %s is not known.\n" % (name[0],))
-
-        if not screen.predict:
-            return
-
         d = ScreenDisplayable(screen, None, None, _widget_properties, scope)
         d.cache = cache_get(screen, _args, kwargs)
         d.update()
@@ -1200,15 +1275,16 @@ def predict_screen(_screen_name, *_args, **kwargs):
 
         renpy.display.predict.displayable(d)
 
-    except:
-        if renpy.config.debug_image_cache:
+    except Exception:
+        if renpy.config.debug_prediction:
             import traceback
 
             print("While predicting screen", _screen_name)
             traceback.print_exc()
+            print()
 
     finally:
-        del scope["_scope"]
+        scope.pop("_scope", None)
 
     renpy.ui.reset()
 
@@ -1263,20 +1339,31 @@ def use_screen(_screen_name, *_args, **kwargs):
     try:
         screen.function(**scope)
     finally:
-        del scope["_scope"]
+        scope.pop("_scope", None)
 
     _current_screen.old_transfers = old_transfers
 
 
-def current_screen():
+def current_screen(): # type: () -> ScreenDisplayable|None
+    """
+    :doc: screens
+    :name: renpy.current_screen
+
+    Returns the ScreendDisplayable corresponding to the screen currently being
+    updated, rendered, or processing an
+
+    See :func:`get_screen` for documented fields on ScreenDisplayable.
+    """
+
     return _current_screen
 
 
-def get_widget(screen, id, layer=None):  # @ReservedAssignment
+def get_displayable(screen, id, layer=None, base=False): # @ReservedAssignment
     """
     :doc: screens
+    :name: renpy.get_displayable
 
-    From the `screen` on `layer`, returns the widget with
+    From the `screen` on `layer`, returns the displayable with
     `id`. Returns None if the screen doesn't exist, or there is no
     widget with that id on the screen.
     """
@@ -1298,15 +1385,23 @@ def get_widget(screen, id, layer=None):  # @ReservedAssignment
     if screen.child is None:
         screen.update()
 
-    rv = screen.widgets.get(id, None)
+    if base:
+        rv = screen.base_widgets.get(id, None)
+    else:
+        rv = screen.widgets.get(id, None)
+
     return rv
 
 
-def get_widget_properties(id, screen=None, layer=None):  # @ReservedAssignment
+get_widget = get_displayable
+
+
+def get_displayable_properties(id, screen=None, layer=None): # @ReservedAssignment
     """
     :doc: screens
+    :name: renpy.get_displayable_properties
 
-    Returns the properties for the widget with `id` in the `screen`
+    Returns the properties for the displayable with `id` in the `screen`
     on `layer`. If `screen` is None, returns the properties for the
     current screen. This can be used from Python or property code inside
     a screen.
@@ -1334,6 +1429,9 @@ def get_widget_properties(id, screen=None, layer=None):  # @ReservedAssignment
     return rv
 
 
+get_widget_properties = get_displayable_properties
+
+
 def before_restart():
     """
     This is called before Ren'Py restarts to put the screens into restart
@@ -1341,7 +1439,7 @@ def before_restart():
     longer defined.
     """
 
-    for k, layer in renpy.display.interface.old_scene.iteritems():
+    for k, layer in renpy.display.interface.old_scene.items():
         if k is None:
             continue
 
@@ -1376,6 +1474,11 @@ def show_overlay_screens(suppress_overlay):
             if get_screen(i) is not None:
                 hide_screen(i)
 
+    for i in renpy.config.always_shown_screens:
+        if get_screen(i) is None:
+            show_screen(i)
+
+
 
 def per_frame():
     """
