@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,19 +19,21 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
-import renpy.display
-import renpy.text
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *
+
+
 import codecs
 import time
 import re
 import sys
 import collections
 import textwrap
+import builtins
 
-import __builtin__
+import renpy
 
-python_builtins = set(dir(__builtin__))
+python_builtins = set(dir(builtins))
 renpy_builtins = set()
 
 image_prefixes = None
@@ -51,6 +53,13 @@ image_prefixes = None
 # The node the report will be about:
 report_node = None
 
+# Collect define/default statements to check for duplication
+all_define_statments = {}
+all_default_statements = {}
+
+# True if at east one error was reported, false otherwise.
+error_reported = False
+
 # Reports a message to the user.
 
 
@@ -61,8 +70,11 @@ def report(msg, *args):
         out = ""
 
     out += msg % args
-    print()
-    print(out.encode('utf-8'))
+    print("")
+    print(out)
+
+    global error_reported
+    error_reported = True
 
 
 added = { }
@@ -71,14 +83,33 @@ added = { }
 # occurs.
 
 
-def add(msg):
+def add(msg, *args):
     if not msg in added:
         added[msg] = True
-        print(unicode(msg).encode('utf-8'))
+        msg = str(msg) % args
+        print(msg)
 
 
-# Trys to evaluate an expression, announcing an error if it fails.
+# Tries to evaluate an expression, announcing an error if it fails.
 def try_eval(where, expr, additional=None):
+    """
+    :doc: lint
+
+    Tries to evaluate an expression, and writes an error to lint.txt if
+    it fails.
+
+    `where`
+        A string giving the location the expression is found. Used to
+        generate an error message of the form "Could not evaluate `expr`
+        in `where`."
+
+    `expr`
+        The expression to try evaluating.
+
+    `additional`
+        If given, an additional line of information that is addded to the
+        error message.
+    """
 
     # Make sure the expression compiles.
     try_compile(where, expr)
@@ -105,10 +136,28 @@ def try_eval(where, expr, additional=None):
 
 
 def try_compile(where, expr, additional=None):
+    """
+    :doc: lint
+
+    Tries to compile an expression, and writes an error to lint.txt if
+    it fails.
+
+    `where`
+        A string giving the location the expression is found. Used to
+        generate an error message of the form "Could not evaluate `expr`
+        in `where`."
+
+    `expr`
+        The expression to try compiling.
+
+    `additional`
+        If given, an additional line of information that is addded to the
+        error message.
+    """
 
     try:
         renpy.python.py_compile_eval_bytecode(expr)
-    except:
+    except Exception:
         report("'%s' could not be compiled as a python expression, %s.", expr, where)
         if additional:
             add(additional)
@@ -202,10 +251,11 @@ def image_exists_precise(name):
 
             try:
                 da = renpy.display.core.DisplayableArguments()
-                da.name=( im[0], ) + tuple(i for i in name[1:] if i in attrs)
-                da.args=tuple(i for i in name[1:] if i in rest)
+                da.name = (im[0],) + tuple(i for i in name[1:] if i in attrs)
+                da.args = tuple(i for i in name[1:] if i in rest)
+                da.lint = True
                 d._duplicate(da)
-            except:
+            except Exception:
                 continue
 
         precise_cache.add(name)
@@ -222,6 +272,11 @@ def image_exists(name, expression, tag, precise=True):
     Checks a scene or show statement for image existence.
     """
 
+    orig = name
+    f = renpy.config.adjust_attributes.get(name[0], None) or renpy.config.adjust_attributes.get(None, None)
+    if f is not None:
+        name = f(name)
+
     # Add the tag to the set of known tags.
     tag = tag or name[0]
     image_prefixes[tag] = True
@@ -238,7 +293,7 @@ def image_exists(name, expression, tag, precise=True):
     if image_exists_precise(name):
         return
 
-    report("'{}' is not an image.".format(" ".join(name)) )
+    report("'%s' is not an image.", " ".join(orig))
 
 
 # Only check each file once.
@@ -274,7 +329,7 @@ def check_displayable(what, d):
     try:
         if isinstance(d, renpy.display.core.Displayable):
             d.visit_all(lambda a: a.predict_one())
-    except:
+    except Exception:
         pass
 
     for fn in files:
@@ -291,7 +346,7 @@ def check_image(node):
 
 def imspec(t):
     if len(t) == 3:
-        return t[0], None, None, t[1], t[2], 0
+        return t[0], None, None, t[1], t[2], 0, None
     if len(t) == 6:
         return t[0], t[1], t[2], t[3], t[4], t[5], None
     else:
@@ -326,9 +381,9 @@ def precheck_show(node):
     tag = imspec(node.imspec)[2]
     image_prefixes[tag] = True
 
-
 # Lints ast.Hide.
 
+
 def check_hide(node):
 
     name, _expression, tag, _at_list, layer, _zorder, _behind = imspec(node.imspec)
@@ -361,14 +416,25 @@ def check_user(node):
 
     try:
         node.get_next()
-    except:
+    except Exception:
         report("Didn't properly report what the next statement should be.")
 
 
+def quote_text(s):
+
+    s = s.replace("\\", "\\\\")
+    s = s.replace("\"", "\\\"")
+    s = s.replace("\t", "\\t")
+    s = s.replace("\n", "\\n")
+
+    return "\"" + s + "\""
+
+
 def text_checks(s):
+
     msg = renpy.text.extras.check_text_tags(s)
     if msg:
-        report("%s (in %s)", msg, repr(s)[1:])
+        report("%s (in %s)", msg, quote_text(s))
 
     if "%" in s and renpy.config.old_substitutions:
 
@@ -395,7 +461,7 @@ def text_checks(s):
                 elif c in "diouxXeEfFgGcrs%":
                     state = 0
                 else:
-                    report("Unknown string format code '%s' (in %s)", fmt, repr(s)[1:])
+                    report("Unknown string format code '%s' (in %s)", fmt, quote_text(s))
                     state = 0
 
             # In a mapping key.
@@ -405,7 +471,7 @@ def text_checks(s):
                     state = 1
 
         if state != 0:
-            report("Unterminated string format code '%s' (in %s)", fmt, repr(s)[1:])
+            report("Unterminated string format code '%s' (in %s)", fmt, quote_text(s))
 
 
 def check_say(node):
@@ -413,10 +479,12 @@ def check_say(node):
     if node.who:
         try:
             char = renpy.ast.eval_who(node.who)
-        except:
+        except Exception:
             report("Could not evaluate '%s' in the who part of a say statement.", node.who)
             add("Perhaps you forgot to define a character?")
             char = None
+    else:
+        char = None
 
     if node.with_:
         try_eval("the with clause of a say statement", node.with_, "Perhaps you forgot to declare, or misspelled, a transition?")
@@ -433,21 +501,22 @@ def check_say(node):
     if not isinstance(char, renpy.character.ADVCharacter):
         return
 
-    if node.attributes is None:
-        return
-
     if char.image_tag is None:
         return
 
-    name = (char.image_tag,) + node.attributes
+    for attributes in (node.attributes, node.temporary_attributes):
+        if attributes is None:
+            continue
 
-    if image_exists_imprecise(name):
-        return
+        name = (char.image_tag,) + attributes
 
-    if image_exists_imprecise(('side', ) + name):
-        return
+        if image_exists_imprecise(name):
+            continue
 
-    report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
+        if image_exists_imprecise(('side',) + name):
+            continue
+
+        report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
 
 
 def check_menu(node):
@@ -488,12 +557,16 @@ def check_while(node):
 
 
 def check_if(node):
-
     for condition, _block in node.entries:
         try_compile("in a condition of the if statement", condition)
 
 
 def check_define(node, kind):
+
+    if node.store == "store.persistent" and kind == "define":
+        report("Define should not be used with a persistent variable. Use default persistent.%s = ... instead.", node.varname)
+        return
+
     if node.store != 'store':
         return
 
@@ -507,6 +580,47 @@ def check_define(node, kind):
         report("'%s %s' replaces a Ren'Py built-in name, which may cause problems.", kind, node.varname)
 
 
+def check_redefined(node, kind):
+    """
+    Check if a define or default statement has already been created.
+    """
+    if kind == 'default':
+        scanned = all_default_statements
+    elif kind == 'define':
+        scanned = all_define_statments
+
+        if not (node.operator == "=" and node.index is None):
+            return
+    else:
+        return
+
+    # Combine store name and varname
+
+    store_name = node.store
+    if store_name.startswith("store."):
+        store_name = store_name[6:]
+
+    if store_name:
+        full_name = "{}.{}".format(store_name, node.varname)
+    else:
+        full_name = node.varname
+
+    if full_name in renpy.config.lint_ignore_redefine:
+        return
+
+    original_node = scanned.get(full_name)
+    if original_node:
+        report(
+            "{} {} already defined at {}:{}".format(
+                kind,
+                full_name,
+                original_node.filename,
+                original_node.linenumber,
+            )
+        )
+    scanned[full_name] = node
+
+
 def check_style_property_displayable(name, property, d):
 
     if not d._duplicatable:
@@ -519,7 +633,7 @@ def check_style_property_displayable(name, property, d):
 
     def sort_short(l):
         l = list(l)
-        l.sort(key=lambda a: len(a))
+        l.sort(key=len)
         return l
 
     alts = sort_short(renpy.style.prefix_alts)
@@ -544,19 +658,18 @@ def check_style_property_displayable(name, property, d):
 def check_style(name, s):
 
     for p in s.properties:
-        for k, v in p.iteritems():
+        for k, v in p.items():
 
             # Treat font specially.
             if k.endswith("font"):
                 if isinstance(v, renpy.text.font.FontGroup):
-                    for f in v.fonts:
+                    for f in set(v.map.values()):
                         check_file(name, f)
                 else:
                     check_file(name, v)
 
             if isinstance(v, renpy.display.core.Displayable):
                 check_style_property_displayable(name, k, v)
-#                check_displayable(kname, v)
 
 
 def check_label(node):
@@ -572,14 +685,22 @@ def check_label(node):
 
     if pi is not None:
 
-        for i in pi.positional:
+        for i, _ in pi.parameters:
             add_arg(i)
+
         add_arg(pi.extrapos)
         add_arg(pi.extrakw)
 
 
+def check_screen(node):
+
+    if (node.screen.parameters is None) and renpy.config.lint_screens_without_parameters:
+        report("The screen %s has not been given a parameter list.", node.screen.name)
+        add("This can be fixed by writing 'screen %s():' instead.", node.screen.name)
+
+
 def check_styles():
-    for full_name, s in renpy.style.styles.iteritems():  # @UndefinedVariable
+    for full_name, s in renpy.style.styles.items(): # @UndefinedVariable
         name = "style." + full_name[0]
         for i in full_name[1:]:
             name += "[{!r}]".format(i)
@@ -610,7 +731,7 @@ def check_filename_encodings():
         try:
             filename.encode("ascii")
             continue
-        except:
+        except Exception:
             pass
 
         report("%s contains non-ASCII characters in its filename.", filename)
@@ -650,6 +771,37 @@ def common(n):
     else:
         return False
 
+def report_character_stats(charastats):
+    """
+    Returns a list of character stat lines.
+    """
+
+    # Keep all the statistics in a list, so that it gets wrapped ionto a
+    rv = [ "Character statistics (for default language):" ]
+
+    count_to_char = collections.defaultdict(list)
+
+    for char, count in charastats.items():
+        count_to_char[count].append(char)
+
+    for count, chars in sorted(count_to_char.items(), reverse=True):
+        chars.sort()
+
+        if len(chars) == 1:
+            start = chars[0] + " has "
+        elif len(chars) == 2:
+            start = chars[0] + " and " + chars[1] + " have "
+        else:
+            start = ", ".join(chars[:-1]) + ", and " + chars[-1] + " have "
+
+        rv.append(
+            " * " + start + humanize(count) +
+            (" block " if count == 1 else " blocks ") + "of dialogue" +
+            (" each." if len(chars) > 1 else ".")
+            )
+
+    return rv
+
 
 def lint():
     """
@@ -659,17 +811,23 @@ def lint():
 
     ap = renpy.arguments.ArgumentParser(description="Checks the script for errors and prints script statistics.", require_command=False)
     ap.add_argument("filename", nargs='?', action="store", help="The file to write to.")
+    ap.add_argument("--error-code", action="store_true", help="If given, the error code is 0 if the game has no lint errros, 1 if lint errors are found.")
 
     args = ap.parse_args()
 
     if args.filename:
-        f = open(args.filename, "w")
+        f = open(args.filename, "w", encoding="utf-8")
         sys.stdout = f
 
     renpy.game.lint = True
 
-    print(codecs.BOM_UTF8)
-    print(unicode(renpy.version + " lint report, generated at: " + time.ctime()).encode("utf-8"))
+    print("\ufeff" + renpy.version + " lint report, generated at: " + time.ctime())
+
+    # Populate default statement values.
+    renpy.exports.execute_default_statement(True)
+
+    # Initialise store and values set by start callbacks.
+    renpy.exports.call_in_new_context('_start_store')
 
     # This supports check_hide.
     global image_prefixes
@@ -681,12 +839,14 @@ def lint():
     # Iterate through every statement in the program, processing
     # them. We sort them in filename, linenumber order.
 
-    all_stmts = [ (i.filename, i.linenumber, i) for i in renpy.game.script.all_stmts ]
-    all_stmts.sort()
+    all_stmts = list(renpy.game.script.all_stmts)
+    all_stmts.sort(key=lambda n : n.filename)
 
     # The current count.
     counts = collections.defaultdict(Count)
 
+    charastats = collections.defaultdict(int)
+
     # The current language.
     language = None
 
@@ -696,11 +856,11 @@ def lint():
 
     global report_node
 
-    for _fn, _ln, node in all_stmts:
+    for node in all_stmts:
         if isinstance(node, (renpy.ast.Show, renpy.ast.Scene)):
             precheck_show(node)
 
-    for _fn, _ln, node in all_stmts:
+    for node in all_stmts:
 
         if common(node):
             continue
@@ -727,6 +887,8 @@ def lint():
             check_say(node)
 
             counts[language].add(node.what)
+            if language is None:
+                charastats[node.who if node.who else 'narrator' ] += 1
 
         elif isinstance(node, renpy.ast.Menu):
             check_menu(node)
@@ -758,12 +920,15 @@ def lint():
 
         elif isinstance(node, renpy.ast.Screen):
             screen_count += 1
+            check_screen(node)
 
         elif isinstance(node, renpy.ast.Define):
             check_define(node, "define")
+            check_redefined(node, "define")
 
         elif isinstance(node, renpy.ast.Default):
             check_define(node, "default")
+            check_redefined(node, "default")
 
     report_node = None
 
@@ -798,34 +963,55 @@ characters per block. """.format(
 
         lines.append(s)
 
-    print()
-    print()
+    print("")
+    print("")
     print("Statistics:")
-    print()
+    print("")
 
     languages = list(counts)
-    languages.sort()
+    languages.sort(key=lambda a : "" if not a else a)
     for i in languages:
         report_language(i)
 
     lines.append("The game contains {0} menus, {1} images, and {2} screens.".format(
         humanize(menu_count), humanize(image_count), humanize(screen_count)))
 
-    for l in lines:
-        for ll in textwrap.wrap(l, 78):
-            print(ll.encode("utf-8"))
+    if renpy.config.developer and renpy.config.lint_character_statistics:
+        lines.append(report_character_stats(charastats))
 
-        print()
+    # Format the lines and lists of lines.
+    for l in lines:
+        if not isinstance(l, (tuple, list)):
+            l = (l,)
+
+        for ll in l:
+
+            if ll.startswith(" * "):
+                prefix = " * "
+                altprefix = "   "
+                ll = ll[3:]
+            else:
+                prefix  = ""
+                altprefix = ""
+
+            for lll in textwrap.wrap(ll, 78 - len(prefix)):
+                print(prefix + lll)
+                prefix = altprefix
+
+        print("")
 
     for i in renpy.config.lint_stats_callbacks:
         i()
 
-    print()
+    print("")
     if renpy.config.developer and (renpy.config.original_developer != "auto"):
         print("Remember to set config.developer to False before releasing.")
-        print()
+        print("")
 
     print("Lint is not a substitute for thorough testing. Remember to update Ren'Py")
     print("before releasing. New releases fix bugs and improve compatibility.")
 
+    if error_reported and args.error_code:
+        renpy.exports.quit(status=1)
+
     return False
